(function (GCN) {

	'use strict';

	/**
	 * @private
	 * @const
	 * @type {string}
	 */
	var GCN_REPOSITORY_ID = 'com.gentics.aloha.GCN.Page';

	/**
	 * @private
	 * @const
	 * @type {object.<string, boolean>} Default page settings.
	 */
	var DEFAULT_SETTINGS = {
		// Load folder information
		folder: true,

		// Lock page when loading it
		update: true
	};

	/**
	 * Searches for the an Aloha editable object of the given id.
	 *
	 * @static
	 * @param {string} id Id of Aloha.Editable object to find.
	 * @return {Aloha.Editable=} The editable object, if wound; otherwise null.
	 */
	function getAlohaEditableById(id) {
		var Aloha = (typeof window !== 'undefined') && window.Aloha;

		if (!Aloha) {
			return null;
		}


		var editables = Aloha.editables;
		var j = editables.length;

		while (j) {
			if (editables[--j].getId() === id) {
				return editables[j];
			}
		}

		return null;
	}

	/**
	 * Checks whether the given tag is a magic link block.
	 *
	 * @private
	 * @static
	 * @param {GCN.Tag} tag Must be a tag that has already been fetched.
	 * @param {object} constructs Set of constructs.
	 * @return {boolean} True if the given tag has the magic link constructId.
	 */
	function isMagicLinkTag(tag, constructs) {
		return !!(constructs[GCN.settings.MAGIC_LINK] &&
		          (constructs[GCN.settings.MAGIC_LINK].constructId ===
			       tag.prop('constructId')));
	}

	/**
	 * Checks whether or not the given block has a corresponding element in the
	 * document DOM.
	 *
	 * @private
	 * @static
	 * @param {object}
	 * @return {boolean} True if an inline element for this block exists.
	 */
	function hasInlineElement(block) {
		return 0 < jQuery('#' + block.element).length;
	}

	/**
	 * @private
	 * @const
	 * @type {number}
	 */
	//var TYPE_ID = 10007;

	/**
	 * @private
	 * @const
	 * @type {Enum}
	 */
	var STATUS = {

		// page was not found in the database
		NOTFOUND: -1,

		// page is locally modified and not yet (re-)published
		MODIFIED: 0,

		// page is marked to be published (dirty)
		TOPUBLISH: 1,

		// page is published and online
		PUBLISHED: 2,

		// Page is offline
		OFFLINE: 3,

		// Page is in the queue (publishing of the page needs to be affirmed)
		QUEUE: 4,

		// page is in timemanagement and outside of the defined timespan
		// (currently offline)
		TIMEMANAGEMENT: 5,

		// page is to be published at a given time (not yet)
		TOPUBLISH_AT: 6
	};

	/**
	 * Given a link, will read the data-gentics-aloha-object-id attribute and
	 * form it, will determine the backend objec that was selected by the
	 * repository browser.
	 *
	 * @param {jQuery.<HTMLElement>} link A link in an editable.
	 * @return {number} The id of the object linked to.
	 */
	function getRepositoryLinkObjectId(link) {
		var data = link.attr('data-gentics-aloha-object-id');

		if (!data) {
			return null;
		}

		var id = data.split('.');

		if (id.length !== 2) {
			return data;
		}

		return id[1] && parseInt(id[1], 10);
	}

	/**
	* @class
	* @name PageAPI
	* @extends ContentObjectAPI
	* @extends TagContainerAPI
	*
	* Page object information can be extened using the default REST-API.
	* options:
	*
	* - update: true
	* Whether the page should be locked in the backend when loading it.
	* default: true
	*
	* - template: true
	* Whether the template information should be embedded in the page object.
	* default: true
	*
	* - folder: true,
	* Whether the folder information should be embedded in the page object.
	* default: true
	* WARNING: do not turn this option off - it will leave the API in a broken
	*          state.
	*
	* - langvars: false,
	* When the language variants shall be embedded in the page response.
	* default: false

	* - workflow: false,
	* When the workflow information shall be embedded in the page response.
	* default: false

	* - pagevars: false,
	* When the page variants shall be embedded in the page response. Page
	* variants will contain folder information.
	* default: false
	*
	* - translationstatus: false
	* Will return information on the page's translation status.
	* default: false
	*/
	var PageAPI = GCN.defineChainback({
		/** @lends PageAPI */

		__chainbacktype__: 'PageAPI',
		_extends: [ GCN.ContentObjectAPI, GCN.TagContainerAPI ],
		_type: 'page',

		/**
		 * @private
		 * @type {Array.<object>} A hash set of block tags belonging to this
		 *                        content object.  This set is added to when
		 *                        this page's tags are rendered.
		 */
		_blocks: {},

		/**
		 * @private
		 * @type {Array.<object>} A hash set of editable tags belonging to this
		 *                        content object.  This set is added to when
		 *                        this page's tags are rendered.
		 */
		_editables: {},

		/**
		 * @type {Array.string} Writable properties for the page object.
		 */
		WRITEABLE_PROPS: ['cdate',
		                  'description',
		                  'fileName',
		                  'folderId', // @TODO Check if moving a page is
		                              //       implemented correctly.
		                  'name',
		                  'priority',
		                  'templateId'],

		/**
		 * Gets all blocks in this page.  Will return an array of all block
		 * objects found in the page AFTER they have been rendered using an
		 * `edit()' call for a contenttag.
		 * NOTE: If you have just loaded the page and not used the `edit()'
		 *       method for any tag the array will be empty.  Only those blocks
		 *       that have been initialized using `edit()' will be available.
		 *
		 * @return {Array.<object>} Array of block objects.
		 */
		'!blocks': function () {
			return this._blocks;
		},

		/**
		 * Looks for a block with the given id in the `_blocks' array.
		 *
		 * @private
		 * @param {string} id The block's id.
		 * @return {?object} The block data object.
		 */
		'!_getBlockById': function (id) {
			return this._blocks[id];
		},

		/**
		 * Maps the received editables into this content object's `_editable'
		 * hash.
		 *
		 * @private
		 * @param {Array.<object>} editables An set of object representing
		 *                                   editable tags that have been
		 *                                   rendered.
		 */
		'!_storeRenderedEditables': function (editables) {
			if (!this.hasOwnProperty('_editables')) {
				this._editables = {};
			}

			var j = editables && editables.length;

			while (j) {
				this._editables[editables[--j].element] = editables[j];
			}
		},

		/**
		 * Maps received blocks of this content object into the `_blocks' hash.
		 *
		 * @private
		 * @param {Array.<object>} blocks An set of object representing
		 *                                block tags that have been rendered.
		 */
		'!_storeRenderedBlocks': function (blocks) {
			if (!this.hasOwnProperty('_blocks')) {
				this._blocks = {};
			}

			var j = blocks && blocks.length;

			while (j) {
				this._blocks[blocks[--j].element] = blocks[j];
			}
		},

		/**
		 * Processes rendered tags, and update the `_blocks' and `_editables'
		 * array accordingly.  This function is called during pre-saving to
		 * update this page's editable tags.
		 *
		 * @private
		 */
		'!_prepareTagsForSaving': function (success, error) {
			if (!this.hasOwnProperty('_deletedBlocks')) {
				this._deletedBlocks = [];
			}

			var that = this;

			this._addNewLinkBlocks(function () {
				that.node().constructs(function (constructs) {
					var id;
					var blocks = [];
					for (id in that._blocks) {
						if (that._blocks.hasOwnProperty(id)) {
							blocks.push(that._blocks[id]);
						}
					}

					that._removeOldLinkBlocks(blocks, constructs, function () {
						that._removeUnusedLinkBlocks(blocks, constructs, function () {
							that._updateEditableBlocks();
							success();
						}, error);
					}, error);
				}, error);
			}, error);
		},

		/**
		 * Removes any link blocks that existed in rendered tags, but have
		 * since been removed by the user while editing.
		 *
		 * @private
		 * @param {Array.<object>} blocks An array of blocks belonging to this
		 *                                page.
		 * @param {object} constrcts A set of constructs.
		 * @param {function} success
		 * @param {function(GCNError):boolean=} error Optional custom error
		 *                                            handler.
		 */
		'!_removeUnusedLinkBlocks': function (blocks, constructs, success,
		                                      error) {
			if (0 === blocks.length) {
				if (success) {
					success();
				}

				return;
			}

			var j = blocks.length;
			var numToProcess = j;

			var onProcess = function () {
				if (0 === --numToProcess) {
					if (success) {
						success();
					}
				}
			};

			var onError = function (error) {
				if (error) {
					error();
				}

				return;
			};

			var that = this;
			var createBlockTagProcessor = function (block) {
				return function (tag) {
					if (!isMagicLinkTag(tag, constructs) &&
							!hasInlineElement(block)) {
						that._deletedBlocks.push(block);
					}

					onProcess();
				};
			};

			while (j) {
				this.tag(blocks[--j].tagname,
					createBlockTagProcessor(blocks[j]), onError);
			}
		},

		/**
		 * Adds any newly created link blocks into this page object.  This is
		 * done by looking for all link blocks that do not have corresponding
		 * tag in this object's `_blocks' list.  For each anchor tag we find,
		 * create a tag for it and, add it in the list of tags.
		 *
		 * @private
		 * @param {function} success Function to invoke if this function
		 *                           successeds.
		 * @param {function(GCNError):boolean=} error Optional custom error
		 *                                            handler.
		 */
		'!_addNewLinkBlocks': function (success, error) {
			var selector = [
				'a[data-gentics-aloha-repository="com.gentics.aloha.GCN.Page"]',
				'a[data-GENTICS-aloha-repository="com.gentics.aloha.GCN.Page"]',
				'a[data-gentics-gcn-url]'
			].join(',');

			var links = jQuery(selector);

			if (0 === links.length) {
				if (success) {
					success();
				}

				return;
			}

			var link;
			var j = links.length;
			var numToProcess = j;

			var onProcessed = function () {
				if (0 === --numToProcess) {
					success();
				}
			};

			var onError = function () {
				if (error) {
					error();
				}

				return;
			};

			var createOnEditHandler = function (link) {
				return function (html, tag) {
					link.attr('id', jQuery(html).attr('id'));
					tag.part('url', getRepositoryLinkObjectId(link));
					onProcessed();
				};
			};

			var tag;

			while (j) {
				link = links.eq(--j);
				if (link.attr('data-gcnignore') === true) {
					onProcessed();
				} else if (this._getBlockById(link.attr('id'))) {
					tag = this.tag(this._getBlockById(link.attr('id')).tagname);
					tag.part('text', link.html());
					tag.part('url', getRepositoryLinkObjectId(link));
					onProcessed();
				} else {
					this.createTag(GCN.settings.MAGIC_LINK, link.html())
					    .edit(createOnEditHandler(link), onError);
				}
			}
		},

		/**
		 * Any links that change from internal GCN links to external links will
		 * have their corresponding blocks added to the '_deletedBlocks' list
		 * since they these links no longer need to be tracked.  Any tags in
		 * this list will be removed during saving.
		 *
		 * @private
		 * @param {Array.<object>} blocks An array of blocks belonging to this
		 *                                page.
		 * @param {object} constrcts A set of constructs.
		 * @param {function} success
		 * @param {function(GCNError):boolean=} error Optional custom error
		 *                                            handler.
		 */
		'!_removeOldLinkBlocks': function (blocks, constructs, success, error) {
			if (0 === blocks.length) {
				if (success) {
					success();
				}

				return;
			}

			var j = blocks.length;
			var numToProcess = j;

			var onProcess = function () {
				if (0 === --numToProcess) {
					if (success) {
						success();
					}
				}
			};

			var onError = function (error) {
				if (error) {
					error();
				}

				return;
			};

			var that = this;
			var createBlockTagProcessor = function (block) {
				return function (tag) {
					if (!isMagicLinkTag(tag, constructs)) {
						onProcess();
						return;
					}

					var a = jQuery('a[id="' + block.element + '"]');

					if (a.length) {
						var isExternal = (GCN_REPOSITORY_ID !==
							a.attr('data-gentics-aloha-repository')) &&
							!a.attr('data-gentics-gcn-url');

						// An external tag was found.  Stop tracking it and
						// remove it from the list of blocks.
						if (isExternal) {
							a.removeAttr('id');
							that._deletedBlocks.push(block);
							delete that._blocks[block.element];
						}

					// No anchor tag was found for this block.  Add it to the
					// "delete" list.
					} else {
						that._deletedBlocks.push(block);
						delete that._blocks[block.element];
					}

					onProcess();
				};
			};

			while (j) {
				this.tag(blocks[--j].tagname,
					createBlockTagProcessor(blocks[j]), onError);
			}
		},

		/**
		 * Writes the contents of editables back into their corresponding tags.
		 * If a corresponding tag cannot be found for an editable, a new one
		 * will be created for it.
		 *
		 * A reference for each editable tag is then added to the `_shadow'
		 * object in order that the tag will be sent with the save request.
		 *
		 * @private
		 */
		'!_updateEditableBlocks': function () {
			var element;
			var elementId;
			var editable;
			var editables = this._editables;
			var tags = this._data.tags;
			var tagname;
			var html;
			var alohaEditable;

			for (elementId in editables) {
				if (editables.hasOwnProperty(elementId)) {
					editable = editables[elementId];
					element = jQuery('#' + elementId);

					// If this editable has no element that was placed in the
					// DOM, then do not attempt to update it.
					if (0 === element.length) {
						continue;
					}

					tagname = editable.tagname;

					if (!tags[tagname]) {
						tags[tagname] = {
							name       : tagname,
							activate   : true,
							properties : {}
						};
					}

					// If the editable element has been `aloha()'fied, then we
					// need to use `getContents()' from is corresponding
					// Aloha.Editable object in order to get clean HTML.

					alohaEditable = getAlohaEditableById(elementId);

					if (alohaEditable) {
						html = alohaEditable.getContents();
						alohaEditable.setUnmodified();
					} else {
						html = element.html();
					}

					tags[tagname].properties[editable.partname] = {
						stringValue: this.encode(html),
						type: 'RICHTEXT'
					};

					this._update('tags.' + tagname, tags[tagname]);
				}
			}
		},

		/**
		 * @see ContentObjectAPI.!_loadParams
		 */
		'!_loadParams': function () {
			return jQuery.extend(DEFAULT_SETTINGS, this._settings);
		},

		/**
		 * Get this page's template.
		 *
		 * @public
		 * @function
		 * @name template
		 * @memberOf PageAPI
		 * @param {funtion(TemplateAPI)=} success Optional callback to receive
		 *                                        a {@link TemplateAPI} object
		 *                                        as the only argument.
		 * @param {function(GCNError):boolean=} error Optional custom error
		 *                                            handler.
		 * @return {TemplateAPI} This page's parent template.
		 */
		'!template': function (success, error) {
			var id = this._fetched ? this.prop('templateId') : null;
			return this._continue(GCN.TempalteAPI, id, success, error);
		},

		/**
		 * @override
		 * @see ContentObjectAPI._save
		 */
		'!_save': function (settings, success, error) {
			var that = this;
			this._continueWith(function () {
				that._prepareTagsForSaving(function () {
					that._persist(settings, success, error);
				}, error);
			}, error);
		},

		//---------------------------------------------------------------------
		// Surface the tag container methods that are applicable for GCN page
		// objects.
		//---------------------------------------------------------------------

		/**
		 * Creates a tag of a given tagtype in this page.
		 *
		 * Exmaple:
		 * <pre>
		 *	createTag('link', 'http://www.gentics.com', onSuccess, onError);
		 * </pre>
		 * or
		 * <pre>
		 *	createTag('link', onSuccess, onError);
		 * </pre>
		 *
		 * @public
		 * @function
		 * @name createTag
		 * @memberOf PageAPI
		 * @param {string|number} construct The name of the construct on which
		 *                                  the tag to be created should be
		 *                                  derived from.  Or the id of that
		 * @param {string=} magicValue Optional property that will override the
		 *                             default values of this tag type.
		 * @param {function(TagAPI)=} success Optional callback that will
		 *                                    receive the newly created tag as
		 *                                    its only argument.
		 * @param {function(GCNError):boolean=} error Optional custom error
		 *                                            handler.
		 * @return {TagAPI} The newly created tag.
		 * @throws INVALID_ARGUMENTS
		 */
		'!createTag': function () {
			return this._createTag.apply(this, arguments);
		},

		/**
		 * Deletes the specified tag from this page.
		 *
		 * @public
		 * @function
		 * @name removeTag
		 * @memberOf PageAPI
		 * @param {string} id The id of the tag to be deleted.
		 * @param {function(PageAPI)=} success Optional callback that receive
		 *                                     this object as its only
		 *                                     argument.
		 * @param {function(GCNError):boolean=} error Optional custom error
		 *                                            handler.
		 */
		removeTag: function () {
			this._removeTag.apply(this, arguments);
		},

		/**
		 * Deletes a set of tags from this page.
		 *
		 * @public
		 * @function
		 * @name removeTags
		 * @memberOf PageAPI
		 * @param {Array.<string>} ids The ids of the set of tags to be
		 *                             deleted.
		 * @param {function(PageAPI)=} success Optional callback that receive
		 *                                     this object as its only
		 *                                     argument.
		 * @param {function(GCNError):boolean=} error Optional custom error
		 *                                            handler.
		 */
		removeTags: function () {
			this._removeTags.apply(this, arguments);
		},

		/**
		 * Marks the page as to be taken offline. This method will change the
		 * state of the page object.
		 *
		 * @public
		 * @function
		 * @name takeOffline
		 * @memberOf PageAPI
		 * @param {funtion(PageAPI)=} success Optional callback to receive this
		 *                                    page object as the only argument.
		 * @param {function(GCNError):boolean=} error Optional custom error
		 *                                            handler.
		 */
		takeOffline: function (success, error) {
			var that = this;

			this._read(function () {
				that._update('status', STATUS.OFFLINE, error);
				if (success) {
					that._save(null, success, error);
				}
			}, error);
		},

		/**
		 * Trigger publish process for the page.
		 *
		 * @public
		 * @function
		 * @name publish
		 * @memberOf PageAPI
		 * @param {funtion(PageAPI)=} success Optional callback to receive this
		 *                                    page object as the only argument.
		 * @param {function(GCNError):boolean=} error Optional custom error
		 *                                            handler.
		 */
		publish: function (success, error) {
			var that = this;
			var parent = this._ancestor();

			var ajax = function () {
				that._authAjax({
					url: GCN.settings.BACKEND_PATH + '/rest/' + that._type +
					     '/publish/' + that.id(),
					type: 'POST',
					json: {}, // there needs to be at least empty content
					          // because of a bug in Jersey
					success: function (response) {
						that._data.status = STATUS.PUBLISHED;
						if (success) {
							success(that);
						}
					},
					error: error
				});
			};

			// If this chainback object has a ancestor, then invoke that
			// parent's `_read()' method before fetching the data for this
			// chainback object.
			if (parent) {
				parent._read(ajax, error);
			} else {
				ajax();
			}
		},

		/**
		 * Renders a preview of the current page.
		 *
		 * @public
		 * @function
		 * @name preview
		 * @memberOf PageAPI
		 * @param {function(string, PageAPI)} success Callback to receive the
		 *                                            rendered page preview as
		 *                                            the first argument, and
		 *                                            this page object as the
		 *                                            second.
		 * @param {function(GCNError):boolean=} error Optional custom error
		 *                                            handler.
		 */
		preview: function (success, error) {
			var that = this;

			this._read(function () {
				that._authAjax({
					url: GCN.settings.BACKEND_PATH + '/rest/' + that._type +
					     '/preview/',
					json: {
						page: that._data // @FIXME Shouldn't this a be merge of
						                 //        the `_shadow' object and the
										 //        `_data'.
					},
					type: 'POST',
					error: error,
					success: function (response) {
						if (success) {
							GCN._handleContentRendered(response.preview,
								function (html) {
									success(html, that);
								});
						}
					}
				});
			}, error);
		},

		/**
		 * Unlocks the page when finishing editing
		 *
		 * @public
		 * @function
		 * @name unlock
		 * @memberOf PageAPI
		 * @param {funtion(PageAPI)=} success Optional callback to receive this
		 *                                    page object as the only argument.
		 * @param {function(GCNError):boolean=} error Optional custom error
		 *                                            handler.
		 */
		unlock: function (success, error) {
			var that = this;
			var parent = this._ancestor();

			var ajax = function () {
				that._authAjax({
					url: GCN.settings.BACKEND_PATH + '/rest/' + that._type +
					     '/cancel/' + that.id(),
					type: 'POST',
					json: {}, // There needs to be at least empty content
					          // because of a bug in Jersey.
					error: error,
					success: function (response) {
						if (success) {
							success(that);
						}
					}
				});
			};

			// If this chainback object has a ancestor, then invoke that
			// parent's `_read()' method before fetching the data for this
			// chainback object.
			if (parent) {
				parent._read(ajax, error);
			} else {
				ajax();
			}
		},

		/**
		 * @see GCN.ContentObjectAPI._processResponse
		 */
		'!_processResponse': function (data) {
			jQuery.extend(this._data, data[this._type]);

			// if data contains page variants turn them into page objects
			if (this._data.pageVariants) {
				var pagevars = [];
				var i;
				for (i = 0; i < this._data.pageVariants.length; i++) {
					pagevars.push(this._continue(GCN.PageAPI,
							this._data.pageVariants[i]));
				}
				this._data.pageVariants = pagevars;
			}
		}

	});

	GCN.page = GCN.exposeAPI(PageAPI);
	GCN.PageAPI = PageAPI;

}(GCN));
